import yaml
import json
import re
import argparse
from api_linter_101.utils import *
import math
from api_linter_101.logger import Logger
import subprocess
import sys

parser = argparse.ArgumentParser()
parser.add_argument("-s", "--spec", type=str, help="Path to the yaml specification file")
parser.add_argument("-r", "--rule", type=str, help="Path to the rules that specification file must follow")
parser.add_argument("-o", "--output", type=str, help="Output the linting errors possible options, json | yaml")

args = parser.parse_args()

error_logger = Logger()

required_path_config = {
   "pattern": pattern_matching,
   "allowed_values":allowed_values,
   "string_match": string_match,
   "gte":gte,
}

key_name_config = {
    "check_keys_list": check_keys_list,
    "check_keys_pattern": check_keys_pattern,
    "check_keys_value": check_keys_value,
    "check_keys_pattern_inverse": check_keys_pattern_inverse,
    "check_keys_B129": check_keys_B129,
    "check_keys_B157":check_keys_B157
}

property_config = {
    "pattern": pattern_matching_property,
    "is_array": is_array,
    "is_date_time": is_date_time
}

def get_spec(filename):
    yaml_file = None
    with open(filename) as yf:
        try:
            yaml_file = yaml.safe_load(yf)
        except yaml.YAMLError as exc:
            print(exc)
    return yaml_file


def parse_required_path(path, spec, spec_file_name):
    # setting path for yq utility tool
    yq_path = "."+path

    path = path.split(".")
    common = path[:-1]
    last = path[-1]

    tmp_spec = spec

    for item in common:
        tmp_spec = tmp_spec.get(item)
        if not tmp_spec:
            return None, None
    
    tmp_spec = tmp_spec.get(last)
    
    if not tmp_spec:
        return None, None
    
    # get line number by using the yq utility tool
    line_num = subprocess.run(["yq", f"{yq_path} | key | line", spec_file_name], capture_output=True)
    line_num = line_num.stdout.decode().strip()

    return tmp_spec, line_num

# write logs to output file
def write_logs(output_fmt, output_file_name):
    with open(output_file_name+f"-output.{output_fmt}", "w") as of:
        if output_fmt=="json":
            print(error_logger.get_logs_json())
            of.write(error_logger.get_logs_json())
        elif output_fmt == "yaml":
            print(error_logger.get_logs_yaml())
            of.write(error_logger.get_logs_yaml())

# array to know the number of implemented rules
implemented_rules = []
def main(spec_file="sample.yaml", rule_file="rules.json", output_fmt="json"):
    rules = get_rules(rule_file)

    yaml_spec = get_spec(spec_file)

    output_file_name = spec_file.split(".")[0]
    # check internal references
    check_refs(yaml_spec, yaml_spec, "", error_logger)

    for rule in rules:
        # id and severity of the rule
        id = rule.get("id")
        severity = rule.get("severity")

        for path_obj in rule.get("paths"):
            
            required_path = path_obj.get("required_path")
            key_name = path_obj.get("key_name")
            properties = path_obj.get("properties")

            if id not in implemented_rules:
                implemented_rules.append(id)

            # if the path is required log the error and continue
            if required_path and severity == "ERROR":
                parsed_path, line_number = parse_required_path(required_path, yaml_spec, spec_file)
                if not parsed_path:
                    error_logger.log_with_property(required_path, severity, path_obj.get("message"), id)
                    continue
            
            for check_obj in path_obj.get("checks"):
                
                for key, value in check_obj.items():
                    # checking for required path and if they exist
                    if key!="message" and required_path:
                        func = required_path_config[key]
                        parsed_path, line_number = parse_required_path(required_path, yaml_spec, spec_file)
                        # if the path does not exist and severity is warn
                        if not parsed_path and severity == "WARN":
                            continue
                        if not func(value, parsed_path):
                            # log errors on checks
                            error_logger.log_with_line_number(line_number, severity, check_obj.get("message"), id)
                    
                    # checking for property attributes of the yaml file 
                    elif key!="message" and properties:
                        func = property_config[key]
                        # spec, line_spec, message, severity, logger, id, pattern
                        func(yaml_spec, "", check_obj.get("message"), severity, error_logger, id, value)
                    
                    # checking for the key name 
                    elif key!="message" and key_name:
                        func = key_name_config[key]
                        func(yaml_spec, "", check_obj.get("message"), severity, error_logger, id, key_name, value)
    error_logger.print_and_save_logs()

def run_script(spec_file, rule_file, output_fmt):
    # create a .env file to save global objects
    with open(".env", "w") as ev:
        ev.write(f"SPEC_FILE={args.spec}")

    output_file_name = args.spec.split(".")[0]
    error_logger.set_file_name_and_format(output_file_name, args.output)
    main(spec_file, rule_file, output_fmt)

if __name__ == "__main__":
    run_script(args.spec, args.rule, args.output)