import re
from datetime import datetime
import json
import subprocess
from dotenv import load_dotenv
from pathlib import Path
import os
import sys

# get rules from json file as a dictionary
def get_rules(filename):
    env_path = Path.cwd() / ".env"
    # set environment variables
    load_dotenv(env_path)
    rules = None
    with open(filename) as jf:
        try:
            rules = json.load(jf)
        except json.JSONDecodeError as exc:
            print(exc)
    return rules

def pattern_matching(pattern, property_):
    if re.match(pattern, property_):
        return True
    return False

def allowed_values(all_values, value):
    if value in all_values:
        return True
    return False

def get_line_number_key(yq_path, logger):
    spec_file_name = os.getenv("SPEC_FILE")

    line_num = subprocess.run(["yq", f"{yq_path} | key | line", spec_file_name], capture_output=True)
    
    int_line = line_num.stdout.decode().strip()
    try:
        int_line = int(int_line)
        return int_line
    except:
        # log the line number
        err = line_num.stderr.decode().strip()
        tmp = err.split("line")
        # log the error and exit the program
        if len(tmp) > 1:
            line_num = int(tmp[1].strip().split(":")[0])
            logger.log_invalid_path(line_num, "Use of json structure is not supported")
            logger.print_and_save_logs()
            sys.exit(1)
    

def get_line_number_value(yq_path):
    spec_file_name = os.getenv("SPEC_FILE")
    line_num = subprocess.run(["yq", f"{yq_path} | line", spec_file_name], capture_output=True)
    line_num = line_num.stdout.decode().stripe()
    return line_num

def check_deep_array(spec, yq_path, message, severity, logger, id, pattern):
    for key, value in spec.items():
        tmp = f".{key}"
        if type(value) is dict:
            check_deep_array(value, yq_path+tmp, message, severity, logger, id, pattern)
        elif type(value) is list:
            if key == "oneOf" or key == "allOf" or key=="anyOf":
                continue
            if not re.match(pattern, key):
                message_ = message
                line_num = get_line_number_key(yq_path + tmp, logger)
                logger.log_with_line_number_key(line_num, severity, message_, id, yq_path+tmp)

# fixed to work with yq_path
def is_array(spec, yq_path, message, severity, logger, id, pattern):
    pattern = "^[a-zA-Z][a-zA-Z_]*(s|es)$"
    return check_deep_array(spec, "", message, severity, logger, id, pattern)

def gte(check, spec_value):
    check = int("".join(check.split(".")))

    spec_value = int("".join(spec_value.split(".")))

    if spec_value >= check:
        return True
    return False

# deep match for this too
def convert_utc_to_datetime(time_str):
    """
        change '2023-01-01T14:05:00.000Z' to '2023-01-01 14:05:00.000'
    """
    pattern = "^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}.[0-9]+Z$"
    if re.match(pattern, time_str) is not None:
        time_str = time_str.replace("T", " ").replace("Z", "")
        return datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S.%f")
    return False

def deep_is_date_time(spec, yq_path, message, severity, logger, id, pattern):
    for key, value in spec.items():
        tmp = f".{key}"

        if type(value) is dict:
            deep_is_date_time(value, yq_path+tmp, message, severity, logger, id, pattern)
            continue
        if type(value) is list:
            continue
        
        # check if the string is a date time object
        try:
            time_obj = convert_utc_to_datetime(value)
            if key[-3:] != "_at" and time_obj:
                message_ = message + f" -> {key}"
                line_num = get_line_number_key(yq_path + tmp, logger)
                logger.log_with_line_number_key_and_value(line_num, severity, message_, id, yq_path+tmp, value)
        except:
            continue


# only works for UTC time format
def is_date_time(spec, yq_path, message, severity, logger, id, pattern=None):
    return deep_is_date_time(spec, "", message, severity, logger, id, pattern)

def string_match(check, spec_value):
    if check == spec_value:
        return True
    else:
        return False

def deep_pattern_matching_property(spec, yq_path, message, severity, logger, id, pattern):
    for key, value in spec.items():
        tmp = f".{key}"
        if not re.match(pattern, key):
            message_ = message
            line_num = get_line_number_key(yq_path + tmp, logger)
            logger.log_with_line_number_key(line_num, severity, message_, id, key)
        if type(value) is dict:
            deep_pattern_matching_property(value, yq_path + tmp, message, severity, logger, id, pattern)

# check if the pattern matches the property name
def pattern_matching_property(spec, yq_path, message, severity, logger, id, pattern):
    deep_pattern_matching_property(spec, "", message, severity, logger, id, pattern)

def parse_internal_refs(ref):
    ref = ref.split("/")[1:]
    return ref

# check if refrences exist
def check_spec_path(spec_file, path_list):
    for path in path_list:
        spec_file = spec_file.get(path)
        if not spec_file:
            return False
    return True


def deep_check_array_refs(spec_iter_file, spec_file, yq_path, logger, message):
    for i, item in enumerate(spec_iter_file):
        tmp = f".[{i}]"
        if type(item) is dict:
            deep_check_refs(item, spec_file, yq_path + tmp, logger, message)

def deep_check_refs(spec_iter_file, spec_file, yq_path, logger, message):
    for key, value in spec_iter_file.items():
        tmp = f".{key}"
        if type(value) is dict:
            deep_check_refs(value, spec_file, yq_path + tmp, logger, message)
        elif type(value) is list:
            deep_check_array_refs(value, spec_file, yq_path + tmp, logger, message)
        elif key == "$ref":
            parsed_path = parse_internal_refs(value)
            if not check_spec_path(spec_file, parsed_path):
                line_num = get_line_number_key(yq_path + tmp, logger)
                logger.log_internal_references(line_num, message, yq_path + tmp, value)
            
# checks to see if internal references exist or not
def check_refs(spec_iter_file, spec_file, yq_path, logger):
    message = "Internal reference does not exist"
    deep_check_refs(spec_iter_file, spec_file, "", logger, message)

# check the keys(shallow) of a property, against a list of prebuilt standards
def check_keys_list_log(spec, yq_path, message, severity, logger, id, key, lst):
    for p_key, value in spec.items():
        tmp = f".{p_key}"
        if p_key not in lst:
            line_num = get_line_number_key(yq_path + tmp, logger)
            logger.log_with_line_number_key(line_num, severity, message, id, yq_path+tmp)

def check_keys_list_array(spec, yq_path, message, severity, logger, id, key, lst):
    for idx, item in enumerate(spec):
        tmp = f".[{idx}]"
        if type(item) is dict:
            check_keys_list(item, yq_path + tmp, message, severity, logger, id, key, lst)

def check_keys_list(spec, yq_path, message, severity, logger, id, key, lst):
    for s_key, value in spec.items():
        tmp = f".{s_key}"
        # for rule B122
        if s_key == "components":
            continue
        if s_key == key and type(value) is dict:
            check_keys_list_log(value, yq_path + tmp, message, severity, logger, id, key, lst)
        elif type(value) is dict:
            check_keys_list(value, yq_path + tmp, message, severity, logger, id, key, lst)
        elif type(value) is list:
            check_keys_list_array(value, yq_path + tmp, message, severity, logger, id, key, lst)

# check the keys(deep) of a property against a pattern
def check_keys_pattern_log(spec, yq_path, message, severity, logger, id, key, pattern):
    for p_key, value in spec.items():
        tmp = f".{p_key}"
        if not re.match(pattern, p_key):
            # print(p_key, pattern, id, yq_path.get(p_key+"_LINE_NUMBER"))
            # print("\n\n")
            line_num = get_line_number_key(yq_path + tmp, logger)
            logger.log_with_line_number_key(line_num, severity, message, id, yq_path + tmp)
        if type(value) is dict:
            check_keys_pattern_log(value, yq_path + tmp, message, severity, logger, id, key, pattern)

def check_keys_pattern_array(spec, yq_path, message, severity, logger, id, key, lst):
    for idx, item in enumerate(spec):
        tmp = ".[{idx}]"
        if type(item) is dict:
            check_keys_pattern(item, yq_path + tmp, message, severity, logger, id, key, lst)

def check_keys_pattern(spec, yq_path, message, severity, logger, id, key, pattern):
    for s_key, value in spec.items():
        tmp = f".{s_key}"
        if s_key == key:
            check_keys_pattern_log(value, yq_path + tmp, message, severity, logger, id, key, pattern)
        elif type(value) is dict:
            check_keys_pattern(value, yq_path + tmp, message, severity, logger, id, key, pattern)
        elif type(value) is list:
            check_keys_pattern_array(value, yq_path + tmp, message, severity, logger, id, key, pattern)

# check if the keys(shallow) of the property matches the standard value
def check_keys_value_log(spec, yq_path, message, severity, logger, id, key, pattern):
    for s_key, s_value in spec.items():
        tmp = f".{s_key}"
        if not re.match(pattern, s_key):
            line_num = get_line_number_key(yq_path + tmp, logger)
            logger.log_with_line_number_key(line_num, severity, message, id, yq_path + tmp)

def check_keys_value_array(spec, yq_path, message, severity, logger, id, key, pattern):
    for idx, item in enumerate(spec):
        tmp = f".[{idx}]"
        if type(item) is dict:
            check_keys_value(item, yq_path + tmp, message, severity, logger, id, key, pattern)

def check_keys_value(spec, yq_path, message, severity, logger, id, key, pattern):
    for s_key, s_value in spec.items():
        tmp = f".{s_key}"
        if s_key == key:
            check_keys_value_log(s_value, yq_path + tmp, message, severity, logger, id, key, pattern)
        elif type(s_value) is dict:
            check_keys_value(s_value, yq_path + tmp, message, severity, logger, id, key, pattern)
        elif type(s_value) is list:
            check_keys_value_array(s_value, yq_path + tmp, message, severity, logger, id, key, pattern)

# check if the keys(shallow) of the property match the pattern, log if it matches the pattern
def check_keys_pattern_inverse_log(spec, yq_path, message, severity, logger, id, key, pattern):
    for s_key, s_value in spec.items():
        tmp = f".{s_key}"
        if re.match(pattern, s_key):
            line_num = get_line_number_key(yq_path + tmp, logger)
            logger.log_with_line_number_key(line_num, severity, message, id, yq_path + tmp)

def check_keys_pattern_inverse_array(spec, yq_path, message, severity, logger, id, key, pattern):
    for idx, item in enumerate(spec):
        tmp = f".[{idx}]"
        if type(item) is dict:
            check_keys_pattern_inverse(item, yq_path + tmp, message, severity, logger, id, key, pattern)

def check_keys_pattern_inverse(spec, yq_path, message, severity, logger, id, key, pattern):
    for s_key, s_value in spec.items():
        tmp = f".{s_key}"
        if s_key == key:
            check_keys_pattern_inverse_log(s_value, yq_path + tmp, message, severity, logger, id, key, pattern)
        elif type(s_value) is dict:
            check_keys_pattern_inverse(s_value, yq_path + tmp, message, severity, logger, id, key, pattern)
        elif type(s_value) is list:
            check_keys_pattern_inverse_array(s_value, yq_path + tmp, message, severity, logger, id, key, pattern)

# finds if the value matches at least one pattern in a list
def findOneB129(lst, val):
    # excluding type
    common_field_names = ["id", "created", "modified", "etag", "type"]
    common_flag = False
    # check if the key is a common field type
    for tmp in common_field_names:
        if re.match(tmp, val):
            common_flag = True
    # if the key is not a common field type skip
    if not common_flag:
        return True
    for tmp in lst:
        if re.match(tmp, val):
            return True
    return False

# check the keys(shallow) of a property, against a list of prebuilt set of standards
def check_keys_B129_log(spec, yq_path, message, severity, logger, id, key, lst):
    for p_key, value in spec.items():
        tmp = f".{p_key}"
        if not findOneB129(lst, p_key):
            line_num = get_line_number_key(yq_path + tmp, logger)
            logger.log_with_line_number_key(line_num, severity, message, id, yq_path + tmp)

def check_keys_B129_array(spec, yq_path, message, severity, logger, id, key, lst):
    for idx, item in enumerate(spec):
        tmp = f".[{idx}]"
        if type(item) is dict:
            check_keys_B129(item, yq_path + tmp, message, severity, logger, id, key, lst)

def check_keys_B129(spec, yq_path, message, severity, logger, id, key, lst):
    for s_key, value in spec.items():
        tmp = f".{s_key}"
        if s_key == key and type(value) is dict:
            check_keys_B129_log(value, yq_path + tmp, message, severity, logger, id, key, lst)
        elif type(value) is dict:
            check_keys_B129(value, yq_path + tmp, message, severity, logger, id, key, lst)
        elif type(value) is list:
            check_keys_B129_array(value, yq_path + tmp, message, severity, logger, id, key, lst)


# check and log
def check_keys_B157_log(spec, yq_path, message, severity, logger, id, key, lst):
    if spec[key] not in lst:
        return
    elif spec[key] == "integer":
        lst = ["int32", "int64", "bigint"]
    elif spec[key] == "number":
        lst = ["decimal", "float", "double"]
    tmp = f".{key}"
    line_num = get_line_number_key(yq_path + tmp, logger)
    if not spec.get("format"):
        logger.log_with_line_number_key(line_num, severity, message + f" {spec[key]} is not defined", id, yq_path + tmp)
        return
    if spec.get("format") not in lst:
        logger.log_with_line_number_key_and_value(line_num, severity, message + f" {spec[key]} must be one of {lst}", id, yq_path + tmp, spec.get("format"))

def check_keys_B157_array(spec, yq_path, message, severity, logger, id, key, lst):
    for idx, item in enumerate(spec):
        tmp = f".[{idx}]"
        if type(item) is dict:
            check_keys_B157(item, yq_path + tmp, message, severity, logger, id, key, lst)

def check_keys_B157(spec, yq_path, message, severity, logger, id, key, lst):
    for s_key, value in spec.items():
        tmp = f".{s_key}"
        if s_key == "type":
            # send the parent property, to check for integer/number formats
            check_keys_B157_log(spec, yq_path, message, severity, logger, id, key, lst)
        elif type(value) is dict:
            check_keys_B157(value, yq_path + tmp, message, severity, logger, id, key, lst)
        elif type(value) is list:
            check_keys_B157_array(value, yq_path + tmp, message, severity, logger, id, key, lst)